<?php

namespace Drupal\default_content_deploy;

use Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\Entity\EntityChangedInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Session\AccountSwitcherInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\default_content_deploy\Event\PreSaveEntityEvent;
use Drupal\hal\LinkManager\LinkManagerInterface;
use Rogervila\ArrayDiffMultidimensional;
use Symfony\Component\Serializer\Serializer;

/**
 * A service for handling import of default content.
 *
 * The importContent() method is almost duplicate of
 *   \Drupal\default_content\Importer::importContent with injected code for
 *   content update. We are waiting for better DC code structure in a future.
 */
class Importer {

  use DependencySerializationTrait;
  use StringTranslationTrait;

  /**
   * Deploy manager.
   *
   * @var \Drupal\default_content_deploy\DeployManager
   */
  protected $deployManager;

  /**
   * Scanned files.
   *
   * @var object[]
   */
  private $files;

  /**
   * Directory to import.
   *
   * @var string
   */
  private $folder;

  /**
   * Data to import.
   *
   * @var array
   */
  private $dataToImport = [];

  /**
   * Data to correct.
   *
   * @var array
   */
  private $dataToCorrect = [];

  /**
   * Path aliases to import.
   *
   * @var array
   */
  private $pathAliasesToImport = [];

  /**
   * Is remove changes of an old content.
   *
   * @var bool
   */
  protected $forceOverride;

  /**
   * The Entity repository manager.
   *
   * @var \Drupal\Core\Entity\EntityRepositoryInterface
   */
  protected $entityRepository;

  /**
   * The cache data.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cache;

  /**
   * Memoization for references that have already been discovered.
   *
   * @var array
   */
  protected $discoveredReferences = [];

  /**
   * The serializer service.
   *
   * @var \Symfony\Component\Serializer\Serializer
   */
  protected $serializer;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The link manager service.
   *
   * @var \Drupal\hal\LinkManager\LinkManagerInterface
   */
  protected $linkManager;

  /**
   * The account switcher.
   *
   * @var \Drupal\Core\Session\AccountSwitcherInterface
   */
  protected $accountSwitcher;

  /**
   * DCD Exporter.
   *
   * @var \Drupal\default_content_deploy\Exporter
   */
  protected $exporter;

  /**
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The event dispatcher.
   *
   * @var \Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher
   */
  protected $eventDispatcher;

  /**
   * @var array
   */
  protected $oldEntityIdLookup = [];

  /**
   * @var array
   */
  protected $entityIdLookup = [];

  /**
   * @var array
   */
  protected $newUuids = [];

  /**
   * @var bool
   */
  protected $verbose = FALSE;

  /**
   * Constructs the default content deploy manager.
   *
   * @param \Symfony\Component\Serializer\Serializer $serializer
   *   The serializer service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   * @param \Drupal\hal\LinkManager\LinkManagerInterface $link_manager
   *   The link manager service.
   * @param \Drupal\Core\Session\AccountSwitcherInterface $account_switcher
   *   The account switcher.
   * @param \Drupal\default_content_deploy\DeployManager $deploy_manager
   *   Deploy manager.
   * @param \Drupal\Core\Entity\EntityRepositoryInterface $entity_repository
   *   The Entity repository manager.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The cache data.
   * @param \Drupal\Core\Database\Connection $database
   *   Database connection.
   * @param \Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher $event_dispatcher
   *   The event dispatcher.
   */
  public function __construct(Serializer $serializer, EntityTypeManagerInterface $entity_type_manager, LinkManagerInterface $link_manager, AccountSwitcherInterface $account_switcher, DeployManager $deploy_manager, EntityRepositoryInterface $entity_repository, CacheBackendInterface $cache, Exporter $exporter, Connection $database, ContainerAwareEventDispatcher $event_dispatcher) {
    $this->serializer = $serializer;
    $this->entityTypeManager = $entity_type_manager;
    $this->linkManager = $link_manager;
    $this->accountSwitcher = $account_switcher;
    $this->deployManager = $deploy_manager;
    $this->entityRepository = $entity_repository;
    $this->cache = $cache;
    $this->exporter = $exporter;
    $this->database = $database;
    $this->eventDispatcher = $event_dispatcher;
  }

  /**
   * Is remove changes of an old content.
   *
   * @param bool $is_override
   *
   * @return \Drupal\default_content_deploy\Importer
   */
  public function setForceOverride(bool $is_override) {
    $this->forceOverride = $is_override;
    return $this;
  }

  /**
   * Set directory to import.
   *
   * @param string $folder
   *   The content folder.
   *
   * @return \Drupal\default_content_deploy\Importer
   */
  public function setFolder(string $folder) {
    $this->folder = $folder;
    return $this;
  }

  /**
   * Get directory to import.
   *
   * @return string
   *   The content folder.
   *
   * @throws \Exception
   */
  protected function getFolder() {
    $folder = $this->folder ?: $this->deployManager->getContentFolder();

    if (!isset($folder)) {
      throw new \Exception('Directory for content deploy is not set.');
    }

    return $folder;
  }

  /**
   * Get Imported data result.
   *
   * @return array
   */
  public function getResult() {
    return $this->dataToImport + $this->pathAliasesToImport;
  }

  /**
   * Set directory to import.
   *
   * @param string $folder
   *   The content folder.
   *
   * @return \Drupal\default_content_deploy\Importer
   */
  public function setVerbose(bool $verbose): void {
    $this->verbose = $verbose;
  }


  /**
   * Import data from JSON and create new entities, or update existing.
   *
   * @return $this
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Exception
   */
  public function prepareForImport(): void {
    // @todo remove because of changes in core >= 9.2
    $this->cache->delete('hal:links:relations');
    $this->files = $this->scan($this->getFolder());

    foreach ($this->files as $file) {
      if (!isset($this->dataToImport[$file->uuid]) && !isset($this->pathAliasesToImport[$file->uuid])) {
        $this->decodeFile($file);
      }
    }
  }

  /**
   * Returns a list of file objects.
   *
   * @param string $directory
   *   Absolute path to the directory to search.
   *
   * @return object[]
   *   List of stdClass objects with name and uri properties.
   */
  public function scan($directory) {
    // Use Unix paths regardless of platform, skip dot directories, follow
    // symlinks (to allow extensions to be linked from elsewhere), and return
    // the RecursiveDirectoryIterator instance to have access to getSubPath(),
    // since SplFileInfo does not support relative paths.
    $flags = \FilesystemIterator::UNIX_PATHS;
    $flags |= \FilesystemIterator::SKIP_DOTS;
    $flags |= \FilesystemIterator::CURRENT_AS_SELF;
    $directory_iterator = new \RecursiveDirectoryIterator($directory, $flags);
    $iterator = new \RecursiveIteratorIterator($directory_iterator);
    $files = [];

    /* @var \SplFileInfo $file_info */
    foreach ($iterator as $file_info) {
      // Skip directories and non-json files.
      if ($file_info->isDir() || $file_info->getExtension() !== 'json') {
        continue;
      }

      $file = new \stdClass();
      $file->name = $file_info->getFilename();
      $file->uuid = str_replace('.json', '', $file->name);
      $file->uri = $file_info->getPathname();
      $file->entity_type_id = basename(dirname($file->uri));
      $file->forceOverride = $this->forceOverride;

      $files[$file->uri] = $file;
    }

    return $files;
  }

  /**
   * Import to entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function import() {
    // Process files in batches.
    $operations = [];
    $total = count($this->dataToImport) + count($this->dataToCorrect) + count($this->pathAliasesToImport);
    $current = 1;

    if ($total === 0) {
      \Drupal::messenger()->addMessage(t('Nothing to import.'));
      return;
    }

    $context = [
      'entityIdLookup' => [],
      'oldEntityIdLookup' => $this->oldEntityIdLookup,
      'skipCorrection' => [],
      'newUuids' => [],
      'verbose' => $this->verbose,
    ];

    $operations[] = [
      [static::class, 'initializeContext'],
      [$context],
    ];

    foreach ($this->dataToImport as $file) {
      $operations[] = [
        [static::class, 'importFile'],
        [$file, $current++, $total, FALSE],
      ];
    }

    foreach ($this->dataToCorrect as $file) {
      $operations[] = [
        [static::class, 'importFile'],
        [$file, $current++, $total, TRUE],
      ];
    }

    foreach ($this->pathAliasesToImport as $file) {
      $operations[] = [
        [static::class, 'importFile'],
        [$file, $current++, $total, FALSE],
      ];
    }

    $batch = [
      'title' => $this->t('Importing Content'),
      'operations' => $operations,
    ];

    batch_set($batch);
  }

  public static function initializeContext($vars, &$context): void {
    $context = array_merge($context, $vars);
    // Set the start time, so we can access across batch operations.
    if (empty($context['results']['start'])) {
      $context['results']['start'] = microtime(TRUE);
    }
  }

  public static function importFile($file, $current, $total, $correction, &$context): void {
    $importer = \Drupal::service('default_content_deploy.importer');
    $importer->processFile($file, $current, $total, $correction, $context);
  }

  /**
   * Prepare file for import.
   *
   * @param $file
   *   The file object to use for import.
   * @param $current
   *   Indicates progress of the batch operations.
   * @param $total
   *   Total number of batch operations.
   * @param bool $correction
   *   Tbd.
   * @param array &$context
   *   Reference to an array that stores the context of the batch process for status updates.
   *
   * @throws \Exception
   */
  protected function processFile($file, $current, $total, $correction, &$context): void {
    $this->entityIdLookup = &$context['entityIdLookup'];
    $this->oldEntityIdLookup = &$context['oldEntityIdLookup'];
    $this->newUuids = &$context['newUuids'];
    $this->verbose = &$context['verbose'];

    if ($correction && array_key_exists($file->uuid, $context['skipCorrection'] ?? [])) {
      if ($this->verbose) {
        $context['message'] = $this->t('@current of @total (@time), skipped correction of @entity_type', [
          '@current' => $current,
          '@total' => $total,
          '@time' => $this->getElapsedTime($context['results']['start']),
          '@entity_type' => $file->entity_type_id,
        ]);
      }

      unset($context['skipCorrection'][$file->uuid]);

      return;
    }

    if (PHP_SAPI === 'cli') {
      $root_user = $this->entityTypeManager->getStorage('user')->load(1);
      $this->accountSwitcher->switchTo($root_user);
    }

    try {
      $is_new = FALSE;
      $this->decodeFile($file, FALSE);

      /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
      $entity = $this->entityRepository->loadEntityByUuid($file->entity_type_id, $file->uuid);

      if ($entity) {
        // Replace entity ID.
        $file->data[$file->key_id][0]['value'] = $entity->id();

        if (!$file->forceOverride) {
          // Skip if the changed time the same or less in the file.
          if ($entity instanceof EntityChangedInterface) {
            // If an entity was refactored to implement the EntityChangedInterface,
            // older exports don't contain the changed field.
            if (isset($file->data['changed'])) {
              $changed_time_file = 0;
              foreach ($file->data['changed'] as $changed) {
                $changed_time = strtotime($changed['value']);
                if ($changed_time > $changed_time_file) {
                  $changed_time_file = $changed_time;
                }
              }
              $changed_time = $entity->getChangedTimeAcrossTranslations();
              if ($changed_time_file <= $changed_time) {
                $context['entityIdLookup'][$file->uuid] = $entity->id();

                if ($this->verbose) {
                  $context['message'] = $this->t('@current of @total (@time), skipped @entity_type @entity_id, file (@date_file) is not newer than database (@date_db)', [
                    '@current' => $current,
                    '@total' => $total,
                    '@time' => $this->getElapsedTime($context['results']['start']),
                    '@entity_type' => $entity->getEntityTypeId(),
                    '@entity_id' => $entity->id(),
                    '@date_file' => date('Y-m-d H:i:s', $changed_time_file),
                    '@date_db' => date('Y-m-d H:i:s', $changed_time),
                  ]);
                }
                $context['skipCorrection'][$file->uuid] = TRUE;

                return;
              }
            }
          }
          else {
            $this->linkManager->setLinkDomain($this->getLinkDomain($file));
            $current_entity_decoded = $this->serializer->decode($this->exporter->getSerializedContent($entity), 'hal_json');
            $diff = ArrayDiffMultidimensional::looseComparison($file->data, $current_entity_decoded);
            if (!$diff) {
              $context['entityIdLookup'][$file->uuid] = $entity->id();

              if ($this->verbose) {
                $context['message'] = $this->t('@current of @total (@time), skipped @entity_type @entity_id, no changes compared to database', [
                  '@current' => $current,
                  '@total' => $total,
                  '@time' => $this->getElapsedTime($context['results']['start']),
                  '@entity_type' => $entity->getEntityTypeId(),
                  '@entity_id' => $entity->id(),
                ]);
              }

              $context['skipCorrection'][$file->uuid] = TRUE;

              return;
            }
          }
        }

        // @todo is this still needed?
        $this->linkManager->setLinkDomain(FALSE);
      }
      else {
        $is_new = TRUE;

        // Ignore ID for creating a new entity.
        unset($file->data[$file->key_id]);
      }

      // All entities with entity references will be imported two times to ensure
      // that all entity references are present and valid. Path aliases will be
      // imported last to have a chance to rewrite them to the new ids of newly
      // created entities.

      $this->linkManager->setLinkDomain($this->getLinkDomain($file));
      $class = $this->entityTypeManager->getDefinition($file->entity_type_id)
        ->getClass();
      if (!$this->preDenormalize($file)) {
        $context['skipCorrection'][$file->uuid] = TRUE;
      }

      /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
      $entity = $this->serializer->denormalize($file->data, $class, 'hal_json', ['request_method' => 'POST']);
      $this->eventDispatcher->dispatch(new PreSaveEntityEvent($entity, $file->data));
      $entity->enforceIsNew($is_new);
      $entity->save();
      $context['entityIdLookup'][$entity->uuid()] = $entity->id();

      if ($is_new) {
        $context['newUuids'][] = $entity->uuid();
      }

      if ($entity->getEntityTypeId() === 'user') {
        // Workaround: store the hashed password directly in the database
        // and avoid the entity API which doesn't provide support for
        // setting password hashes directly.
        $hashed_pass = $file->data['pass'][0]['value'] ?? FALSE;
        if ($hashed_pass) {
          $this->database->update('users_field_data')
            ->fields([
              'pass' => $hashed_pass,
            ])
            ->condition('uid', $entity->id(), '=')
            ->execute();
        }
      }

      // @todo is this still needed?
      $this->linkManager->setLinkDomain(FALSE);

      if (PHP_SAPI === 'cli') {
        $this->accountSwitcher->switchBack();
      }

      if ($this->verbose) {
        $context['message'] = $this->t('@current of @total (@time), @operation @entity_type @entity_id', [
          '@current' => $current,
          '@total' => $total,
          '@time' => $this->getElapsedTime($context['results']['start']),
          '@operation' => $is_new ? $this->t('created') : $this->t('updated'),
          '@entity_type' => $entity->getEntityTypeId(),
          '@entity_id' => $entity->id(),
        ]);
      }
    }
    catch (\Exception $e) {
      $context['message'] = $this->t('@current of @total (@time), error on importing @entity_type @uuid: @message', [
        '@current' => $current,
        '@total' => $total,
        '@time' => $this->getElapsedTime($context['results']['start']),
        '@entity_type' => $file->entity_type_id,
        '@uuid' => $file->uuid,
        '@message' => $e->getMessage(),
      ]);
    }
  }

  /**
   * Calculates and formats the elapsed time.
   *
   * @param float $start
   *   The start time of the overall batch process.
   *
   * @return string
   *   The formatted elapsed time in minutes.
   */
  public function getElapsedTime($start) {
    $end = microtime(TRUE);
    $diff = $end - $start;
    $elapsed_time = number_format($diff / 60, 2) . ' ' . $this->t('minutes');

    return $elapsed_time;
  }

  /**
   * Gets url from file for set to Link manager.
   *
   * @param $file
   */
  protected function getLinkDomain($file) {
    $link = $file->data['_links']['type']['href'];
    $url_data = parse_url($link);
    $host = "{$url_data['scheme']}://{$url_data['host']}";
    return (!isset($url_data['port'])) ? $host : "{$host}:{$url_data['port']}";
  }

  /**
   * Prepare file to import.
   *
   * @param $file
   *
   * @return $this
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Exception
   */
  protected function decodeFile($file, $add_to_import = TRUE): void {
    // Check that this file has not been decoded already.
    if (array_key_exists($file->name, $this->discoveredReferences)) {
      return;
    }

    // Get parsed data.
    $parsed_data = file_get_contents($file->uri);

    // Decode.
    try {
      $decode = $this->serializer->decode($parsed_data, 'hal_json');
    }
    catch (\Exception $e) {
      throw new \RuntimeException(sprintf('Unable to decode %s', $file->uri), $e->getCode(), $e);
    }

    if ($add_to_import) {
      $references = $this->getReferences($decode);

      // Record that we have checked references of current file.
      $this->discoveredReferences[$file->name] = $file;
      if ($references) {
        foreach ($references as $reference) {
          $this->decodeFile($reference, $add_to_import);
        }
      }
    }

    // Prepare data for import.
    $file->data = $decode;
    if (!isset($file->references)) {
      $file->references = array_keys($references ?? []);
    }

    $this->preAddToImport($file);
    if ($add_to_import) {
      $this->addToImport($file);
    }
  }

  /**
   * Here we can edit data`s value before importing.
   *
   * @param $file
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   */
  protected function preAddToImport($file): void {
    $entity_type_object = $this->entityTypeManager->getDefinition($file->entity_type_id);

    // Keys of entity.
    $file->key_id = $entity_type_object->getKey('id');

    // Some old exports don't have the entity ID.
    if (isset($file->data[$file->key_id][0]['value'])) {
      $this->oldEntityIdLookup[$file->entity_type_id][$file->data[$file->key_id][0]['value']][] = $file->uuid;
    }

    // @see path_entity_base_field_info().
    // @todo offer an event to let third party modules register their content
    //       types.
    if (in_array($file->entity_type_id, ['taxonomy_term', 'node', 'media'])) {
      unset($file->data['path']);
    }

    // Ignore revision and id of entity.
    if ($key_revision_id = $entity_type_object->getKey('revision')) {
      unset($file->data[$key_revision_id]);
    }
  }

  /**
   * This event is triggered before decoding to an entity.
   *
   * @param $file
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function preDenormalize($file): bool {
    $needs_second_run = FALSE;

    $this->updateTargetRevisionId($file->data);
    $needs_second_run |= $this->updateInternalLinks($file->data);

    if ($file->entity_type_id === 'path_alias') {
      $this->updatePathAliasTargetId($file->data);
    }

    return $needs_second_run;
  }

  /**
   * Adding prepared data for import.
   *
   * @param $file
   *
   * @return $this
   */
  protected function addToImport($file) {
    // Save memory.
    unset($file->data);

    if ($file->entity_type_id === 'path_alias') {
      $this->pathAliasesToImport[$file->uuid] = $file;
    }
    else {
      $this->dataToImport[$file->uuid] = $file;
      if (!empty($file->references)) {
        $this->dataToCorrect[$file->uuid] = $file;
      }
    }

    return $this;
  }

  /**
   * Get all reference by entity array content.
   *
   * @param array $content
   *
   * @return array
   */
  private function getReferences(array $content) {
    $references = [];

    if (isset($content['_embedded'])) {
      foreach ($content['_embedded'] as $link) {
        foreach ($link as $reference) {
          if ($reference) {
            if ($uuid = $reference['uuid'][0]['value'] ?? NULL) {
              if ($path = $this->getPathToFileByName($uuid)) {
                $references[$uuid] = $this->files[$path];
              }
            }
          }
        }
      }
    }

    return $references;
  }

  /**
   * Get path to file by Name.
   *
   * @param $name
   *
   * @return false|int|string
   */
  private function getPathToFileByName($name) {
    $array_column = array_column($this->files, 'name', 'uri');
    return array_search($name . '.json', $array_column);
  }

  /**
   * Get Entity type ID by link.
   *
   * @param $link
   *
   * @return string|string[]
   */
  private function getEntityTypeByLink($link): string {
    $type = $this->linkManager->getTypeInternalIds($link);

    if ($type) {
      $entity_type_id = $type['entity_type'];
    }
    else {
      $components = array_reverse(explode('/', $link));
      $entity_type_id = $components[1];
      // @todo remove this line when core is >= 9.2
      $this->cache->invalidate('hal:links:types');
    }

    return $entity_type_id;
  }

  /**
   * If this entity contains a reference field with target revision is value,
   * we should to update it.
   *
   * @param $decode
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function updateTargetRevisionId(&$decode): void {
    if (isset($decode['_embedded'])) {
      foreach ($decode['_embedded'] as $link_key => $link) {
        if (array_column($link, 'target_revision_id')) {
          foreach ($link as $ref_key => $reference) {
            $url = $reference['_links']['type']['href'];
            $uuid = $reference['uuid'][0]['value'];
            $entity_type = $this->getEntityTypeByLink($url);
            $entity = $this->entityRepository->loadEntityByUuid($entity_type, $uuid);

            // Update the Target revision id if child entity exist on this site.
            if ($entity) {
              $revision_id = $entity->getRevisionId();
              $decode['_embedded'][$link_key][$ref_key]['target_revision_id'] = $revision_id;
            }
          }
        }
      }
    }
  }

  /**
   * Rewrite internal links to target entity IDs that were assigned during
   * import.
   *
   * @param array $decode
   *   The decoded entity.
   *
   * @return bool
   *   TRUE if the entity contains an internal link field.
   */
  private function updateInternalLinks(array &$decode): bool {
    $has_internal_links = FALSE;

    foreach ($decode as $field => $items) {
      foreach ($items as $index => $item) {
        foreach ($item as $name => $value) {
          if ('uri' === $name) {
            if (str_starts_with($value, 'internal:')) {
              $decode[$field][$index][$name] = 'internal:' . $this->getUpdatedInternalPath(str_replace('internal:', '', $value));
              $has_internal_links = TRUE;
            }
            elseif (str_starts_with($value, 'entity:')) {
              $decode[$field][$index][$name] = 'entity:' . trim($this->getUpdatedInternalPath(str_replace('entity:', '/', $value), $item['target_uuid'] ?? NULL), '/');
              $has_internal_links = TRUE;
            }
          }
        }
      }
    }

    return $has_internal_links;
  }

  /**
   * Rewrite path aliases to target entity IDs that were assigned during import.
   *
   * @param array $decode
   *   The decoded entity.
   */
  private function updatePathAliasTargetId(array &$decode) {
    if ($path = $decode['path'][0]['value'] ?? NULL) {
      $decode['path'][0]['value'] = $this->getUpdatedInternalPath($path);
    }
  }

  /**
   * Rewrite a path to target entity IDs that were assigned during import.
   *
   * @param string $path
   *   The path from exported content.
   *
   * @return string
   *   The updated path.
   *
   * @throws \InvalidArgumentException
   */
  private function getUpdatedInternalPath($path, $uuid = NULL): string {
    if (preg_match('@^/(\w+)/(\d+)([/?#].*|)$@', $path, $matches)) {
      $entity_type_id = str_replace('_', '/', $matches[1]);
      if (!$uuid) {
        $uuids = $this->oldEntityIdLookup[$entity_type_id][$matches[2]] ?? [];
        if (count($uuids) > 1) {
          $new_uuid = FALSE;
          foreach ($this->oldEntityIdLookup[$entity_type_id][$matches[2]] as $uuid_candidate) {
            if (in_array($uuid_candidate, $this->newUuids, TRUE)) {
              // If an internal path that needs to be updated or created targets
              // an entity ID of new entity, this one is preferred.
              $uuid = $uuid_candidate;
              $new_uuid = TRUE;
              break;
            }
            $new_uuid = FALSE;
          }
          if (!$new_uuid) {
            throw new \InvalidArgumentException(sprintf('Multiple UUIDs (%s) share the same entity ID %d.', implode(', ', $this->oldEntityIdLookup[$entity_type_id][$matches[2]]), $matches[2]));
          }
        }
        elseif (!empty($uuids)) {
          $uuid = reset($uuids);
        }
      }
      if ($uuid) {
        if ($id = $this->entityIdLookup[$uuid] ?? NULL) {
          return '/' . $matches[1] . '/' . $id . $matches[3];
        }
      }
    }

    return $path;
  }

}
